@@ -20,7 +20,9 @@
 #include <osmocom/core/tdef.h>
 #include <osmocom/core/time_cc.h>
 #include <osmocom/core/linuxlist.h>
+#include <osmocom/core/hashtable.h>
 #include <osmocom/core/linuxrbtree.h>
+#include <osmocom/core/utils.h>
 
 #include <osmocom/crypt/auth.h>
 
@@ -96,6 +98,13 @@ struct smlc_config;
 #define DLCI2RSL_LINK_ID(dlci) \
 	((dlci & 0xc0) == 0xc0 ? 0x40 : 0x00) | (dlci & 0x07)
 
+typedef uint16_t gsm_bts_nr_t; /* See (struct gsm_bts *)->nr */
+/* BTS_NR_MAX = ((2 << ((sizeof(gsm_bts_nr_t) * 8) - 1)) - 1)
+ * This is needed as a constant so that the value can be stringified properly: */
+#define BTS_NR_MAX 65535
+#define BTS_NR_MAX_STR OSMO_STRINGIFY_VAL(BTS_NR_MAX)
+#define BTS_NR_VTY_ARG_VAL "<0-" BTS_NR_MAX_STR ">"
+
 /* 3-bit long values */
 #define EARFCN_PRIO_INVALID 8
 #define EARFCN_MEAS_BW_INVALID 8
@@ -463,6 +472,8 @@ struct gsm_subscriber_connection {
 		struct gsm_subscriber_connection *call;
 		/* Cell (BTS) FSM */
 		struct osmo_fsm_inst *fi;
+		/* BTS to notify about ongoing call */
+		struct gsm_bts *notify_bts;
 		/* lchan to be assigned */
 		struct gsm_lchan *new_lchan;
 		/* MGW peer */
@@ -881,7 +892,7 @@ extern struct osmo_tdef_group bsc_tdef_group[];
 
 struct gsm_network *gsm_network_init(void *ctx);
 
-struct gsm_bts *gsm_bts_num(const struct gsm_network *net, int num);
+struct gsm_bts *gsm_bts_num(const struct gsm_network *net, gsm_bts_nr_t num);
 struct gsm_bts *gsm_bts_by_cell_id(const struct gsm_network *net,
 				   const struct gsm0808_cell_id *cell_id,
 				   int match_idx);
@@ -975,9 +986,14 @@ struct gsm_network {
 	struct rate_ctr_group *bsc_ctrs;
 	struct osmo_stat_item_group *bsc_statg;
 
-	unsigned int num_bts;
+	gsm_bts_nr_t num_bts;
 	struct llist_head bts_list;
 	struct llist_head bts_rejected;
+	DECLARE_HASHTABLE(bts_by_nr, 10);
+	DECLARE_HASHTABLE(bts_by_lac, 10);
+	DECLARE_HASHTABLE(bts_by_lac_ci, 10);
+#define LAC_CI_HASHTABLE_KEY(lac, ci) ((((uint32_t)(ci)) << sizeof(lac)) | (uint32_t)(lac))
+	DECLARE_HASHTABLE(bts_by_ci, 10);
 
 	/* BTS-based counters when we can't find the actual BTS
 	 * e.g. when conn->lchan is NULL */
@@ -1085,8 +1101,6 @@ extern void talloc_ctx_init(void *ctx_root);
 
 enum gsm_bts_type parse_btstype(const char *arg);
 const char *btstype2str(enum gsm_bts_type type);
-struct gsm_bts *gsm_bts_by_lac(struct gsm_network *net, unsigned int lac,
-				struct gsm_bts *start_bts);
 
 extern void *tall_bsc_ctx;
 
